
/*--------------------------------------------------------------------*/
/*--- Volcheck: The volatile-checking tool.              vc_main.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Volcheck, a Valgrind tool for checking accesses
   to volatile storage.

   Copyright (C) 2007, 2008 Eric Eide
      eeide@cs.utah.edu

   This file is derived from `none/nl_main.c' in the Valgrind software.
   That file is:

   Copyright (C) 2002-2007 Nicholas Nethercote
      njn@valgrind.org

   Much of this file is derived from `lackey/lk_main.c' in the Valgrind
   software.  That file is:

   Copyright (C) 2002-2007 Nicholas Nethercote
      njn@valgrind.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

//
// ENE, Nov 11 2007: below, a big donation of code from `lackey/lk_main.c'.
//

// This tool shows how to do some basic instrumentation.
//
// There are three kinds of instrumentation it can do.  They can be turned
// on/off independently with command line options:
//
// * --basic-counts   : do basic counts, eg. number of instructions
//                      executed, jumps executed, etc.
// * --detailed-counts: do more detailed counts:  number of loads, stores
//                      and ALU operations of different sizes.
// * --trace-mem      : trace all (data) memory accesses.
//
// The code for each kind of instrumentation is guarded by a clo_* variable:
// clo_basic_counts, clo_detailed_counts and clo_trace_mem.
//
// If you want to modify any of the instrumentation code, look for the code
// that is guarded by the relevant clo_* variable (eg. clo_trace_mem)
// If you're not interested in the other kinds of instrumentation you can
// remove them.  If you want to do more complex modifications, please read
// VEX/pub/libvex_ir.h to understand the intermediate representation.
//
//
// Specific Details about --trace-mem=yes
// --------------------------------------
// Lackey's --trace-mem code is a good starting point for building Valgrind
// tools that act on memory loads and stores.  It also could be used as is,
// with its output used as input to a post-mortem processing step.  However,
// because memory traces can be very large, online analysis is generally
// better.
//
// It prints memory data access traces that look like this:
//
//   instr   : 0x0023C790, 2  # instruction read at 0x0023C790 of size 2
//   instr   : 0x0023C792, 5
//     store : 0xBE80199C, 4  # data store at 0xBE80199C of size 4
//   instr   : 0x0025242B, 3
//     load  : 0xBE801950, 4  # data load at 0xBE801950 of size 4
//   instr   : 0x0023D476, 7
//     modify: 0x0025747C, 1  # data modify at 0x0025747C of size 1
//   instr   : 0x0023DC20, 2
//     load  : 0x00254962, 1
//     load  : 0xBE801FB3, 1
//   instr   : 0x00252305, 1
//     load  : 0x00254AEB, 1
//     store : 0x00257998, 1
//
// Every instruction executed has an "instr" event representing it.
// Instructions that do memory accesses are followed by one or more "load",
// "store" or "modify" events.  Some instructions do more than one load or
// store, as in the last two examples in the above trace.
//
// Here are some examples of x86 instructions that do different combinations
// of loads, stores, and modifies.
//
//    Instruction          Memory accesses                  Event sequence
//    -----------          ---------------                  --------------
//    add %eax, %ebx       No loads or stores               instr
//
//    movl (%eax), %ebx    loads (%eax)                     instr, load
//
//    movl %eax, (%ebx)    stores (%ebx)                    instr, store
//
//    incl (%ecx)          modifies (%ecx)                  instr, modify
//
//    cmpsb                loads (%esi), loads(%edi)        instr, load, load
//
//    call*l (%edx)        loads (%edx), stores -4(%esp)    instr, load, store
//    pushl (%edx)         loads (%edx), stores -4(%esp)    instr, load, store
//    movsw                loads (%esi), stores (%edi)      instr, load, store
//
// Instructions using x86 "rep" prefixes are traced as if they are repeated
// N times.
//
// Lackey with --trace-mem gives good traces, but they are not perfect, for
// the following reasons:
//
// - It does not trace into the OS kernel, so system calls and other kernel
//   operations (eg. some scheduling and signal handling code) are ignored.
//
// - It could model loads and stores done at the system call boundary using
//   the pre_mem_read/post_mem_write events.  For example, if you call
//   fstat() you know that the passed in buffer has been written.  But it
//   currently does not do this.
//
// - Valgrind replaces some code (not much) with its own, notably parts of
//   code for scheduling operations and signal handling.  This code is not
//   traced.
//
// - There is no consideration of virtual-to-physical address mapping.
//   This may not matter for many purposes.
//
// - Valgrind modifies the instruction stream in some very minor ways.  For
//   example, on x86 the bts, btc, btr instructions are incorrectly
//   considered to always touch memory (this is a consequence of these
//   instructions being very difficult to simulate).
//
// - Valgrind tools layout memory differently to normal programs, so the
//   addresses you get will not be typical.  Thus Lackey (and all Valgrind
//   tools) is suitable for getting relative memory traces -- eg. if you
//   want to analyse locality of memory accesses -- but is not good if
//   absolute addresses are important.
//
// Despite all these warnings, Dullard's results should be good enough for a
// wide range of purposes.  For example, Cachegrind shares all the above
// shortcomings and it is still useful.
//
// For further inspiration, you should look at cachegrind/cg_main.c which
// uses the same basic technique for tracing memory accesses, but also groups
// events together for processing into twos and threes so that fewer C calls
// are made and things run faster.

#include "pub_tool_basics.h"
#include "pub_tool_tooliface.h"
#include "pub_tool_libcassert.h"
#include "pub_tool_libcprint.h"
#include "pub_tool_debuginfo.h"
#include "pub_tool_libcbase.h"
#include "pub_tool_options.h"
#include "pub_tool_machine.h"     // VG_(fnptr_to_fnentry)

/*------------------------------------------------------------*/
/*--- Command line options                                 ---*/
/*------------------------------------------------------------*/

/* Command line options controlling instrumentation kinds, as described at
 * the top of this file. */
static Bool clo_basic_counts      = True;
static Bool clo_detailed_counts   = False;
static Bool clo_trace_mem         = True;

static Bool clo_print_mem_per     = True;
static Bool clo_print_mem_summary = True;
static Bool clo_missing_loc_ok    = False;

/* The name of the function of which the number of calls (under
 * --basic-counts=yes) is to be counted, with default. Override with command
 * line option --fnname. */
static Char* clo_fnname = "_dl_runtime_resolve";

static Bool vc_process_cmd_line_option(Char* arg)
{
   VG_STR_CLO(arg, "--fnname", clo_fnname)
   else VG_BOOL_CLO(arg, "--basic-counts",      clo_basic_counts)
   else VG_BOOL_CLO(arg, "--detailed-counts",   clo_detailed_counts)
   else VG_BOOL_CLO(arg, "--trace-mem",         clo_trace_mem)
   else VG_BOOL_CLO(arg, "--print-mem-per",     clo_print_mem_per)
   else VG_BOOL_CLO(arg, "--print-mem-summary", clo_print_mem_summary)
   else VG_BOOL_CLO(arg, "--missing-loc-ok",    clo_missing_loc_ok)
   else
      return False;
   
   tl_assert(clo_fnname);
   tl_assert(clo_fnname[0]);
   return True;
}

static void vc_print_usage(void)
{  
   VG_(printf)(
"    --basic-counts=no|yes       count instructions, jumps, etc. [yes]\n"
"    --detailed-counts=no|yes    count loads, stores and ALU ops [no]\n"
"    --trace-mem=no|yes          trace all loads and stores [yes]\n"
"    --print-mem-per=no|yes      output a message for every load and store\n"
"                                [yes]\n"
"    --print-mem-summary=no|yes  output a memory access summary [yes]\n"
"    --missing-loc-ok=no|yes     trace loads and stores that occur at points\n"
"                                without source locations [no]\n"
"    --fnname=<name>             count calls to <name> (only used if\n"
"                                --basic-count=yes)  [_dl_runtime_resolve]\n"
   );
}

static void vc_print_debug_usage(void)
{  
   VG_(printf)(
"    (none)\n"
   );
}

/*------------------------------------------------------------*/
/*--- Stuff for --basic-counts                             ---*/
/*------------------------------------------------------------*/

/* Nb: use ULongs because the numbers can get very big */
static ULong n_func_calls    = 0;
static ULong n_SBs_entered   = 0;
static ULong n_SBs_completed = 0;
static ULong n_IRStmts       = 0;
static ULong n_guest_instrs  = 0;
static ULong n_Jccs          = 0;
static ULong n_Jccs_untaken  = 0;
static ULong n_IJccs         = 0;
static ULong n_IJccs_untaken = 0;

static void add_one_func_call(void)
{
   n_func_calls++;
}

static void add_one_SB_entered(void)
{
   n_SBs_entered++;
}

static void add_one_SB_completed(void)
{
   n_SBs_completed++;
}

static void add_one_IRStmt(void)
{
   n_IRStmts++;
}

static void add_one_guest_instr(void)
{
   n_guest_instrs++;
}

static void add_one_Jcc(void)
{
   n_Jccs++;
}

static void add_one_Jcc_untaken(void)
{
   n_Jccs_untaken++;
}

static void add_one_inverted_Jcc(void)
{
   n_IJccs++;
}

static void add_one_inverted_Jcc_untaken(void)
{
   n_IJccs_untaken++;
}

/*------------------------------------------------------------*/
/*--- Stuff for --detailed-counts                          ---*/
/*------------------------------------------------------------*/

/* --- Operations --- */

typedef enum { OpLoad=0, OpStore=1, OpAlu=2 } Op;

#define N_OPS 3


/* --- Types --- */

#define N_TYPES 9

static Int type2index ( IRType ty )
{
   switch (ty) {
      case Ity_I1:      return 0;
      case Ity_I8:      return 1;
      case Ity_I16:     return 2;
      case Ity_I32:     return 3;
      case Ity_I64:     return 4;
      case Ity_I128:    return 5;
      case Ity_F32:     return 6;
      case Ity_F64:     return 7;
      case Ity_V128:    return 8;
      default: tl_assert(0); break;
   }
}

static HChar* nameOfTypeIndex ( IRType ty )
{
   switch (ty) {
      case 0: return "I1";   break;
      case 1: return "I8";   break;
      case 2: return "I16";  break;
      case 3: return "I32";  break;
      case 4: return "I64";  break;
      case 5: return "I128"; break;
      case 6: return "F32";  break;
      case 7: return "F64";  break;
      case 8: return "V128"; break;
      default: tl_assert(0); break;
   }
}


/* --- Counts --- */

static ULong detailCounts[N_OPS][N_TYPES];

/* The helper that is called from the instrumented code. */
static VG_REGPARM(1)
void increment_detail(ULong* detail)
{
   (*detail)++;
}

/* A helper that adds the instrumentation for a detail. */
static void instrument_detail(IRSB* sb, Op op, IRType type)
{
   IRDirty* di;
   IRExpr** argv;
   const UInt typeIx = type2index(type);

   tl_assert(op < N_OPS);
   tl_assert(typeIx < N_TYPES);

   argv = mkIRExprVec_1( mkIRExpr_HWord( (HWord)&detailCounts[op][typeIx] ) );
   di = unsafeIRDirty_0_N( 1, "increment_detail",
                              VG_(fnptr_to_fnentry)( &increment_detail ), 
                              argv);
   addStmtToIRSB( sb, IRStmt_Dirty(di) );
}

/* Summarize and print the details. */
static void print_details ( void )
{
   Int typeIx;
   VG_(message)(Vg_UserMsg,
                "   Type        Loads       Stores       AluOps");
   VG_(message)(Vg_UserMsg,
                "   -------------------------------------------");
   for (typeIx = 0; typeIx < N_TYPES; typeIx++) {
      VG_(message)(Vg_UserMsg,
                   "   %4s %,12llu %,12llu %,12llu", 
                   nameOfTypeIndex( typeIx ),
                   detailCounts[OpLoad ][typeIx],
                   detailCounts[OpStore][typeIx],
                   detailCounts[OpAlu  ][typeIx]
      );
   }
}


/*------------------------------------------------------------*/
/*--- Stuff for --trace-mem                                ---*/
/*------------------------------------------------------------*/

#define MAX_DSIZE    512

typedef
   IRExpr 
   IRAtom;

typedef 
   enum { Event_Ir, Event_Dr, Event_Dw, Event_Dm }
   EventKind;

typedef
   struct {
      EventKind  ekind;
      IRAtom*    addr;
      Int        size;
   }
   Event;

/* Up to this many unnotified events are allowed.  Must be at least two,
   so that reads and writes to the same address can be merged into a modify.
   Beyond that, larger numbers just potentially induce more spilling due to
   extending live ranges of address temporaries. */

/*
 * ENE, Tue Mar 25 2008: try to increase precision via `NO_EVENT_BUFFER'.
 */
#define NO_EVENT_BUFFER 1
#if NO_EVENT_BUFFER
#  define N_EVENTS 1
#else
#  define N_EVENTS 4
#endif

/* Maintain an ordered list of memory events which are outstanding, in
   the sense that no IR has yet been generated to do the relevant
   helper calls.  The BB is scanned top to bottom and memory events
   are added to the end of the list, merging with the most recent
   notified event where possible (Dw immediately following Dr and
   having the same size and EA can be merged).

   This merging is done so that for architectures which have
   load-op-store instructions (x86, amd64), the instr is treated as if
   it makes just one memory reference (a modify), rather than two (a
   read followed by a write at the same address).

   At various points the list will need to be flushed, that is, IR
   generated from it.  That must happen before any possible exit from
   the block (the end, or an IRStmt_Exit).  Flushing also takes place
   when there is no space to add a new event.

   If we require the simulation statistics to be up to date with
   respect to possible memory exceptions, then the list would have to
   be flushed before each memory reference.  That's a pain so we don't
   bother.

   Flushing the list consists of walking it start to end and emitting
   instrumentation IR for each event, in the order in which they
   appear. */

static Event events[N_EVENTS];
static Int   events_used = 0;

static Bool loc_available;
static Char loc_filename[256];
static UInt loc_linenum;

#if 0
/* This is "the future": how to find data symbols in Valgrind 3.4. */
static Char sym_name[256];

static Char*
lookup_symbol(Addr addr)
{
   OffT offset;
   if (VG_(get_datasym_and_offset)(addr, sym_name, 256, &offset)) {
      return sym_name;
   }
   return 0;
}
#endif /* 0 */

/*
 * XXX --- need to cache symbols
 */
static HChar*
lookup_symbol(Addr addr)
{
   SegInfo* seg = VG_(find_seginfo)(addr);
   Int      how_many;
   Int      i;

   if (!seg) {
      return 0;
   }

   how_many = VG_(seginfo_syms_howmany)(seg);
   for (i = 0; i < how_many; ++i) {
      Addr   sym_addr;
      Addr   sym_tocptr;
      UInt   sym_size;
      HChar* sym_name;
    
      VG_(seginfo_syms_getidx)(seg, i,
			       &sym_addr, &sym_tocptr, &sym_size, &sym_name);
      if (sym_addr == addr) {
          return sym_name;
      }
   }
   return 0;
}

/*****/

typedef struct AddrRecord {
   struct AddrRecord *	next;
   
   Addr			addr;
   HChar *		symbol;
   ULong		loads[4]; /* XXX hardcoded ick */
   ULong		stores[4];
   ULong		modifies[4];
} AddrRecord;

#define ADDR_TABLE_SIZE (4096)
static AddrRecord addr_table[ADDR_TABLE_SIZE];
static int addr_table_cnt = 0;
static AddrRecord addr_table_root; /* Root of the linked list in table */

static AddrRecord addr_record_scratch;

static void
init_addr_record(AddrRecord *r)
{
   int i;

   r->addr = 0;
   r->symbol = 0;
   for (i = 0; i < 4; ++i) { r->loads[i] = 0; }
   for (i = 0; i < 4; ++i) { r->stores[i] = 0; }
   for (i = 0; i < 4; ++i) { r->modifies[i] = 0; }
}

static AddrRecord *
alloc_addr_record(void)
{
   AddrRecord *r;

   if (addr_table_cnt >= ADDR_TABLE_SIZE) {
      /*
       * Assert now, to make allocation errors immediately apparent.
       *
       * But: the callers of this function should be prepared to deal with
       * allocation errors at run time.  Remove the assertion if you prefer
       * that behavior.
       */
      tl_assert(!"address table is full");
      return 0;
   }

   r = &(addr_table[addr_table_cnt]);
   ++addr_table_cnt;
   init_addr_record(r);

   return r;
}

static void
init_addr_table(void)
{
   init_addr_record(&addr_table_root);
   addr_table_cnt = 0;
}

static AddrRecord *
find_addr_record(Addr addr)
{
   AddrRecord *p = &addr_table_root;
   AddrRecord *next;
   AddrRecord *new;

   /*
    * Find the record for `addr', or find the place to insert a new record for
    * `addr' into the list.
    */
   while (1) {
      next = p->next;
      if (!next) {
	 break;
      }
      if (next->addr == addr) {
	 return next;
      }
      if (next->addr > addr) {
	 break;
      }
      p = next;
   }

   /*
    * If we get here, then `p' points to the record that should precede the
    * newly allocate record for `addr'.
    */
   new = alloc_addr_record();
   if (!new) {
      return 0;
   }
   new->addr = addr;
   new->symbol = lookup_symbol(addr);
   new->next = next;
   p->next = new;
   return new;
}

static const AddrRecord *
record_event(Addr addr, SizeT size, EventKind kind)
{
   AddrRecord *r = find_addr_record(addr);
   ULong *counts;
   int i;
   
   if (!r) {
      // XXX print out a warning
      init_addr_record(&addr_record_scratch);
      addr_record_scratch.addr = addr;
      addr_record_scratch.symbol = lookup_symbol(addr);

      return &addr_record_scratch;
   }
   
   switch (kind) {
   case Event_Dr:	counts = r->loads;	break;
   case Event_Dw:	counts = r->stores;	break;
   case Event_Dm:	counts = r->modifies;	break;
   default:		tl_assert(0);		break;
   }
   switch (size) {
   case 1:	i = 0;		break;
   case 2:	i = 1;		break;
   case 4:	i = 2;		break;
   case 8:	i = 3;		break;
   default:	tl_assert(0);	break;
   }
   counts[i] += 1;

   return r;
}

/*****/

static const char*
event_name(EventKind kind)
{
   switch (kind) {
   case Event_Ir: return "ifetch"; break;
   case Event_Dr: return "load"; break;
   case Event_Dw: return "store"; break;
   case Event_Dm: return "modify"; break;
   }
   return "unknown";
}

static void
trace_event(Addr addr, SizeT size, EventKind kind)
{
   const AddrRecord *record;
   
   if (!clo_missing_loc_ok) {
      /* Only trace activities within the program source text. */
      if (!loc_available)
	 return;
   }
   /* Only trace globals. */
   if ((VG_(seginfo_sect_kind)(addr) != Vg_SectBSS)
       && (VG_(seginfo_sect_kind)(addr) != Vg_SectData))
     return;

   record = record_event(addr, size, kind);

   if (clo_print_mem_per) {
      HChar *symbol_name = record->symbol;
      if (!symbol_name) {
	 symbol_name = "<unknown>";
      }
      if (loc_available) {
	 VG_(printf)("  %s:%u", loc_filename, loc_linenum);
      } else {
	 VG_(printf)("  ???:???");
      }
      VG_(printf)(":  %s %s    (%08p, %d)\n",
		  event_name(kind), symbol_name, addr, size);
   }
}

static VG_REGPARM(2) void
trace_instr(Addr addr, SizeT size)
{
/*
   VG_(printf)("instr   : %08p, %d\n", addr, size);
*/
  /* Determine where we are from the instruction being loaded. */
   loc_available =
      VG_(get_filename_linenum)(addr,
                                loc_filename, sizeof(loc_filename),
                                0, 0, 0,
                                &loc_linenum);
}

static VG_REGPARM(2) void
trace_load(Addr addr, SizeT size)
{
   trace_event(addr, size, Event_Dr);
}

static VG_REGPARM(2) void
trace_store(Addr addr, SizeT size)
{
   trace_event(addr, size, Event_Dw);
}

static VG_REGPARM(2) void
trace_modify(Addr addr, SizeT size)
{
   trace_event(addr, size, Event_Dm);
}


static void flushEvents(IRSB* sb)
{
   Int        i;
   Char*      helperName;
   void*      helperAddr;
   IRExpr**   argv;
   IRDirty*   di;
   Event*     ev;

   for (i = 0; i < events_used; i++) {

      ev = &events[i];
      
      // Decide on helper fn to call and args to pass it.
      switch (ev->ekind) {
         case Event_Ir: helperName = "trace_instr";
                        helperAddr =  trace_instr;  break;

         case Event_Dr: helperName = "trace_load";
                        helperAddr =  trace_load;   break;

         case Event_Dw: helperName = "trace_store";
                        helperAddr =  trace_store;  break;

         case Event_Dm: helperName = "trace_modify";
                        helperAddr =  trace_modify; break;
         default:
            tl_assert(0);
      }

      if (helperAddr) {
         // Add the helper.
         argv = mkIRExprVec_2( ev->addr, mkIRExpr_HWord( ev->size ) );
         di   = unsafeIRDirty_0_N( /*regparms*/2, 
                                   helperName, VG_(fnptr_to_fnentry)( helperAddr ),
                                   argv );
         addStmtToIRSB( sb, IRStmt_Dirty(di) );
      }
   }

   events_used = 0;
}

// WARNING:  If you aren't interested in instruction reads, you can omit the
// code that adds calls to trace_instr() in flushEvents().  However, you
// must still call this function, addEvent_Ir() -- it is necessary to add
// the Ir events to the events list so that merging of paired load/store
// events into modify events works correctly.
static void addEvent_Ir ( IRSB* sb, IRAtom* iaddr, UInt isize )
{
   Event* evt;
   tl_assert( (VG_MIN_INSTR_SZB <= isize && isize <= VG_MAX_INSTR_SZB)
            || VG_CLREQ_SZB == isize );
   if (events_used == N_EVENTS)
      flushEvents(sb);
   tl_assert(events_used >= 0 && events_used < N_EVENTS);
   evt = &events[events_used];
   evt->ekind = Event_Ir;
   evt->addr  = iaddr;
   evt->size  = isize;
   events_used++;
#if NO_EVENT_BUFFER
   flushEvents(sb);
#endif
}

static
void addEvent_Dr ( IRSB* sb, IRAtom* daddr, Int dsize )
{
   Event* evt;
   tl_assert(isIRAtom(daddr));
   tl_assert(dsize >= 1 && dsize <= MAX_DSIZE);
   if (events_used == N_EVENTS)
      flushEvents(sb);
   tl_assert(events_used >= 0 && events_used < N_EVENTS);
   evt = &events[events_used];
   evt->ekind = Event_Dr;
   evt->addr  = daddr;
   evt->size  = dsize;
   events_used++;
#if NO_EVENT_BUFFER
   flushEvents(sb);
#endif
}

static
void addEvent_Dw ( IRSB* sb, IRAtom* daddr, Int dsize )
{
   Event* lastEvt;
   Event* evt;
   tl_assert(isIRAtom(daddr));
   tl_assert(dsize >= 1 && dsize <= MAX_DSIZE);

   // Is it possible to merge this write with the preceding read?
   lastEvt = &events[events_used-1];
   if (events_used > 0
    && lastEvt->ekind == Event_Dr
    && lastEvt->size  == dsize
    && eqIRAtom(lastEvt->addr, daddr))
   {
      lastEvt->ekind = Event_Dm;
      return;
   }

   // No.  Add as normal.
   if (events_used == N_EVENTS)
      flushEvents(sb);
   tl_assert(events_used >= 0 && events_used < N_EVENTS);
   evt = &events[events_used];
   evt->ekind = Event_Dw;
   evt->size  = dsize;
   evt->addr  = daddr;
   events_used++;
#if NO_EVENT_BUFFER
   flushEvents(sb);
#endif
}


/*------------------------------------------------------------*/
/*--- Basic tool functions                                 ---*/
/*------------------------------------------------------------*/

static void vc_post_clo_init(void)
{
   Int op, tyIx;

   if (clo_detailed_counts) {
      for (op = 0; op < N_OPS; op++)
         for (tyIx = 0; tyIx < N_TYPES; tyIx++)
            detailCounts[op][tyIx] = 0;
   }

   init_addr_table();
}

static
IRSB* vc_instrument ( VgCallbackClosure* closure,
                      IRSB* sbIn,
                      VexGuestLayout* layout, 
                      VexGuestExtents* vge,
                      IRType gWordTy, IRType hWordTy )
{
   IRDirty*   di;
   Int        i;
   IRSB*      sbOut;
   Char       fnname[100];
   IRType     type;
   IRTypeEnv* tyenv = sbIn->tyenv;
   Addr       iaddr = 0, dst;
   UInt       ilen = 0;
   Bool       condition_inverted = False;

   if (gWordTy != hWordTy) {
      /* We don't currently support this case. */
      VG_(tool_panic)("host/guest word size mismatch");
   }

   /* Set up SB */
   sbOut = deepCopyIRSBExceptStmts(sbIn);

   // Copy verbatim any IR preamble preceding the first IMark
   i = 0;
   while (i < sbIn->stmts_used && sbIn->stmts[i]->tag != Ist_IMark) {
      addStmtToIRSB( sbOut, sbIn->stmts[i] );
      i++;
   }

   if (clo_basic_counts) {
      /* Count this superblock. */
      di = unsafeIRDirty_0_N( 0, "add_one_SB_entered", 
                                 VG_(fnptr_to_fnentry)( &add_one_SB_entered ),
                                 mkIRExprVec_0() );
      addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
   }

   if (clo_trace_mem) {
      events_used = 0;
   }

   for (/*use current i*/; i < sbIn->stmts_used; i++) {
      IRStmt* st = sbIn->stmts[i];
      if (!st || st->tag == Ist_NoOp) continue;

      if (clo_basic_counts) {
         /* Count one VEX statement. */
         di = unsafeIRDirty_0_N( 0, "add_one_IRStmt", 
                                    VG_(fnptr_to_fnentry)( &add_one_IRStmt ), 
                                    mkIRExprVec_0() );
         addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
      }
      
      switch (st->tag) {
         case Ist_NoOp:
         case Ist_AbiHint:
         case Ist_Put:
         case Ist_PutI:
         case Ist_MBE:
            addStmtToIRSB( sbOut, st );
            break;

         case Ist_IMark:
            if (clo_basic_counts) {
               /* Needed to be able to check for inverted condition in Ist_Exit */
               iaddr = st->Ist.IMark.addr;
               ilen  = st->Ist.IMark.len;

               /* Count guest instruction. */
               di = unsafeIRDirty_0_N( 0, "add_one_guest_instr",
                                          VG_(fnptr_to_fnentry)( &add_one_guest_instr ), 
                                          mkIRExprVec_0() );
               addStmtToIRSB( sbOut, IRStmt_Dirty(di) );

               /* An unconditional branch to a known destination in the
                * guest's instructions can be represented, in the IRSB to
                * instrument, by the VEX statements that are the
                * translation of that known destination. This feature is
                * called 'SB chasing' and can be influenced by command
                * line option --vex-guest-chase-thresh.
                *
                * To get an accurate count of the calls to a specific
                * function, taking SB chasing into account, we need to
                * check for each guest instruction (Ist_IMark) if it is
                * the entry point of a function.
                */
               tl_assert(clo_fnname);
               tl_assert(clo_fnname[0]);
               if (VG_(get_fnname_if_entry)(st->Ist.IMark.addr, 
                                            fnname, sizeof(fnname))
                   && 0 == VG_(strcmp)(fnname, clo_fnname)) {
                  di = unsafeIRDirty_0_N( 
                          0, "add_one_func_call", 
                             VG_(fnptr_to_fnentry)( &add_one_func_call ), 
                             mkIRExprVec_0() );
                  addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
               }
            }
            if (clo_trace_mem) {
               // WARNING: do not remove this function call, even if you
               // aren't interested in instruction reads.  See the comment
               // above the function itself for more detail.
               addEvent_Ir( sbOut, mkIRExpr_HWord( (HWord)st->Ist.IMark.addr ),
                            st->Ist.IMark.len );
            }
            addStmtToIRSB( sbOut, st );
            break;

         case Ist_WrTmp:
            // Add a call to trace_load() if --trace-mem=yes.
            if (clo_trace_mem) {
               IRExpr* data = st->Ist.WrTmp.data;
               if (data->tag == Iex_Load) {
                  addEvent_Dr( sbOut, data->Iex.Load.addr,
                               sizeofIRType(data->Iex.Load.ty) );
               }
            }
            if (clo_detailed_counts) {
               IRExpr* expr = st->Ist.WrTmp.data;
               type = typeOfIRExpr(sbOut->tyenv, expr);
               tl_assert(type != Ity_INVALID);
               switch (expr->tag) {
                  case Iex_Load:
                     instrument_detail( sbOut, OpLoad, type );
                     break;
                  case Iex_Unop:
                  case Iex_Binop:
                  case Iex_Triop:
                  case Iex_Qop:
                  case Iex_Mux0X:
                     instrument_detail( sbOut, OpAlu, type );
                     break;
                  default:
                     break;
               }
            }
            addStmtToIRSB( sbOut, st );
            break;

         case Ist_Store:
            if (clo_trace_mem) {
               IRExpr* data  = st->Ist.Store.data;
               addEvent_Dw( sbOut, st->Ist.Store.addr,
                            sizeofIRType(typeOfIRExpr(tyenv, data)) );
            }
            if (clo_detailed_counts) {
               type = typeOfIRExpr(sbOut->tyenv, st->Ist.Store.data);
               tl_assert(type != Ity_INVALID);
               instrument_detail( sbOut, OpStore, type );
            }
            addStmtToIRSB( sbOut, st );
            break;

         case Ist_Dirty: {
            if (clo_trace_mem) {
               Int      dsize;
               IRDirty* d = st->Ist.Dirty.details;
               if (d->mFx != Ifx_None) {
                  // This dirty helper accesses memory.  Collect the details.
                  tl_assert(d->mAddr != NULL);
                  tl_assert(d->mSize != 0);
                  dsize = d->mSize;
                  if (d->mFx == Ifx_Read || d->mFx == Ifx_Modify)
                     addEvent_Dr( sbOut, d->mAddr, dsize );
                  if (d->mFx == Ifx_Write || d->mFx == Ifx_Modify)
                     addEvent_Dw( sbOut, d->mAddr, dsize );
               } else {
                  tl_assert(d->mAddr == NULL);
                  tl_assert(d->mSize == 0);
               }
            }
            addStmtToIRSB( sbOut, st );
            break;
         }

         case Ist_Exit:
            if (clo_basic_counts) {
               // The condition of a branch was inverted by VEX if a taken
               // branch is in fact a fall trough according to client address
               tl_assert(iaddr != 0);
               dst = (sizeof(Addr) == 4) ? st->Ist.Exit.dst->Ico.U32 :
                                           st->Ist.Exit.dst->Ico.U64;
               condition_inverted = (dst == iaddr + ilen);

               /* Count Jcc */
               if (!condition_inverted)
                  di = unsafeIRDirty_0_N( 0, "add_one_Jcc", 
                                          VG_(fnptr_to_fnentry)( &add_one_Jcc ), 
                                          mkIRExprVec_0() );
               else
                  di = unsafeIRDirty_0_N( 0, "add_one_inverted_Jcc",
                                          VG_(fnptr_to_fnentry)( &add_one_inverted_Jcc ),
                                          mkIRExprVec_0() );

               addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
            }
            if (clo_trace_mem) {
               flushEvents(sbOut);
            }

            addStmtToIRSB( sbOut, st );      // Original statement

            if (clo_basic_counts) {
               /* Count non-taken Jcc */
               if (!condition_inverted)
                  di = unsafeIRDirty_0_N( 0, "add_one_Jcc_untaken", 
                                          VG_(fnptr_to_fnentry)(
                                             &add_one_Jcc_untaken ),
                                          mkIRExprVec_0() );
               else
                  di = unsafeIRDirty_0_N( 0, "add_one_inverted_Jcc_untaken",
                                          VG_(fnptr_to_fnentry)(
                                             &add_one_inverted_Jcc_untaken ),
                                          mkIRExprVec_0() );

               addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
            }
            break;

         default:
            tl_assert(0);
      }
   }

   if (clo_basic_counts) {
      /* Count this basic block. */
      di = unsafeIRDirty_0_N( 0, "add_one_SB_completed", 
                                 VG_(fnptr_to_fnentry)( &add_one_SB_completed ),
                                 mkIRExprVec_0() );
      addStmtToIRSB( sbOut, IRStmt_Dirty(di) );
   }

   if (clo_trace_mem) {
      /* At the end of the sbIn.  Flush outstandings. */
      flushEvents(sbOut);
   }

   return sbOut;
}

static void
print_addr_accesses(const AddrRecord *r, const ULong *cnts,
		    const char *kind)
{
   int i;
   int size;

   for (i = 0, size = 1; i < 4; ++i, size *= 2) {
      ULong n = cnts[i];
      if (n != 0) {
	 VG_(printf)("  %-8s  %08p  %d %llu\n", kind, r->addr, size, n);
      }
   }
}

static void vc_fini(Int exitcode)
{
   char percentify_buf[4]; /* Two digits, '%' and 0. */
   const int percentify_size = sizeof(percentify_buf);
   const int percentify_decs = 0;
   
   tl_assert(clo_fnname);
   tl_assert(clo_fnname[0]);

   VG_(message)(Vg_UserMsg,
		"Volcheck recorded access info for %d unique memory addresses",
		addr_table_cnt);
   if (addr_table_cnt >= ADDR_TABLE_SIZE) {
      VG_(message)(Vg_UserMsg,
		   ("WARNING: Volcheck may have missed some accesses due to "
		    "table exhaustion"));
   }
   if (clo_print_mem_summary) {
      AddrRecord *p;
      VG_(message)(Vg_UserMsg,
		   "Below is a memory-access summary (op, addr, size, count)");
      VG_(message)(Vg_UserMsg, "");
      for (p = addr_table_root.next; p != 0; p = p->next) {
	 print_addr_accesses(p, p->loads, event_name(Event_Dr));
	 print_addr_accesses(p, p->stores, event_name(Event_Dw));
	 print_addr_accesses(p, p->modifies, event_name(Event_Dm));
#if 0
	 VG_(printf)("  %08p", p->addr);
	 if (p->symbol)
	    VG_(printf)(" %s\n", p->symbol);
	 else
	    VG_(printf)(" <unknown>\n");
#endif
      }
   }

   if (clo_basic_counts) {
      ULong total_Jccs = n_Jccs + n_IJccs;
      ULong taken_Jccs = (n_Jccs - n_Jccs_untaken) + n_IJccs_untaken;

      VG_(message)(Vg_UserMsg, "");
      VG_(message)(Vg_UserMsg,
         "Counted %,llu calls to %s()", n_func_calls, clo_fnname);

      VG_(message)(Vg_UserMsg, "Jccs:");
      VG_(message)(Vg_UserMsg, "  total:         %,llu", total_Jccs);
      VG_(percentify)(taken_Jccs, (total_Jccs ? total_Jccs : 1),
         percentify_decs, percentify_size, percentify_buf);
      VG_(message)(Vg_UserMsg, "  taken:         %,llu (%s)", 
         taken_Jccs, percentify_buf);
      
      VG_(message)(Vg_UserMsg, "");
      VG_(message)(Vg_UserMsg, "Executed:");
      VG_(message)(Vg_UserMsg, "  SBs entered:   %,llu", n_SBs_entered);
      VG_(message)(Vg_UserMsg, "  SBs completed: %,llu", n_SBs_completed);
      VG_(message)(Vg_UserMsg, "  guest instrs:  %,llu", n_guest_instrs);
      VG_(message)(Vg_UserMsg, "  IRStmts:       %,llu", n_IRStmts);
      
      VG_(message)(Vg_UserMsg, "");
      VG_(message)(Vg_UserMsg, "Ratios:");
      tl_assert(n_SBs_entered); // Paranoia time.
      VG_(message)(Vg_UserMsg, "  guest instrs : SB entered  = %3u : 10",
         10 * n_guest_instrs / n_SBs_entered);
      VG_(message)(Vg_UserMsg, "       IRStmts : SB entered  = %3u : 10",
         10 * n_IRStmts / n_SBs_entered);
      tl_assert(n_guest_instrs); // Paranoia time.
      VG_(message)(Vg_UserMsg, "       IRStmts : guest instr = %3u : 10",
         10 * n_IRStmts / n_guest_instrs);
   }

   if (clo_detailed_counts) {
      VG_(message)(Vg_UserMsg, "");
      VG_(message)(Vg_UserMsg, "IR-level counts by type:");
      print_details();
   }

   if (clo_basic_counts) {
      VG_(message)(Vg_UserMsg, "");
      VG_(message)(Vg_UserMsg, "Exit code:       %d", exitcode);
   }
}

static void vc_pre_clo_init(void)
{
   VG_(details_name)            ("Volcheck");
   VG_(details_version)         (NULL);
   VG_(details_description)     (
      "a tool for checking the use of volatile storage");
   VG_(details_copyright_author)(
      "Copyright (C) 2007, 2008, and GNU GPL'd, by Eric Eide.");
   VG_(details_bug_reports_to)  ("eeide@cs.utah.edu");
   VG_(details_avg_translation_sizeB) (200);

   VG_(basic_tool_funcs)        (vc_post_clo_init,
                                 vc_instrument,
                                 vc_fini);
   VG_(needs_command_line_options)(vc_process_cmd_line_option,
                                   vc_print_usage,
                                   vc_print_debug_usage);
#if 1
   /* Unnecessary (and does not work) in Valgrind 3.4. */
   VG_(needs_data_syms)();
#endif

   /* No core events to track */

   /* Set the VEX control params.  See `libvex.h'. */
   /*
    * Turn the VEX optimizer off.  *THIS IS ESSENTIAL*.
    *
    * When the VEX optimizer is enabled, it will happily remove instructions
    * that appear to be useless.  For example, it will remove the read of `g_1'
    * in the following code:
    *
    *   volatile int g_1;
    *   void f() { int size = (g_1 && 0); }
    *
    * Such optimizations defeat the entire purpose of this tool.
    */
   VG_(clo_vex_control).iropt_level = 0;
}

VG_DETERMINE_INTERFACE_VERSION(vc_pre_clo_init)

/*--------------------------------------------------------------------*/
/*--- end                                                          ---*/
/*--------------------------------------------------------------------*/
